import argparse
from datetime import datetime, timedelta
import os

from .MonteCarloService import MonteCarloService
from .CsvService import CsvService

def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--FileName", default="")
    parser.add_argument("--Delimeter", default=";")
    parser.add_argument("--ClosedDateColumn", default="Closed Date")
    parser.add_argument("--DateFormat", default="%m/%d/%Y %I:%M:%S %p")
    parser.add_argument("--TargetDate", default="")
    parser.add_argument("--TargetDateFormat", default="%d.%m.%Y")
    parser.add_argument("--RemainingItems", default="10")
    parser.add_argument("--History", default="30")
    parser.add_argument("--SaveCharts", default=False, action=argparse.BooleanOptionalAction)

    return parser.parse_args()

def get_closed_items_history(csv_service, monte_carlo_service, file_name, delimeter, closed_date_column, date_format):    
    work_items = csv_service.get_closed_items(file_name, delimeter, closed_date_column, date_format)
    closed_items_history = monte_carlo_service.create_closed_items_history(work_items)
    return closed_items_history

def print_logo():
    logo = r"""
     /$$                 /$$           /$$$$$$$                           /$$                /$$      /$$                  /$$      
    | $$                | $$          | $$__  $$                         | $$               | $$  /$ | $$                 | $$      
    | $$       /$$$$$$ /$$$$$$        | $$  \ $$/$$$$$$  /$$$$$$  /$$$$$$| $$ /$$$$$$       | $$ /$$$| $$ /$$$$$$  /$$$$$$| $$   /$$
    | $$      /$$__  $|_  $$_/        | $$$$$$$/$$__  $$/$$__  $$/$$__  $| $$/$$__  $$      | $$/$$ $$ $$/$$__  $$/$$__  $| $$  /$$/
    | $$     | $$$$$$$$ | $$          | $$____| $$$$$$$| $$  \ $| $$  \ $| $| $$$$$$$$      | $$$$_  $$$| $$  \ $| $$  \__| $$$$$$/ 
    | $$     | $$_____/ | $$ /$$      | $$    | $$_____| $$  | $| $$  | $| $| $$_____/      | $$$/ \  $$| $$  | $| $$     | $$_  $$ 
    | $$$$$$$|  $$$$$$$ |  $$$$/      | $$    |  $$$$$$|  $$$$$$| $$$$$$$| $|  $$$$$$$      | $$/   \  $|  $$$$$$| $$     | $$ \  $$
    |________/\_______/  \___/        |__/     \_______/\______/| $$____/|__/\_______/      |__/     \__/\______/|__/     |__/  \__/
                                                                | $$                                                                
                                                                | $$                                                                
                                                                |__/                                                                
    """
    print(logo)

def check_if_file_exists(file_path, raise_if_not_found = False):
    if not os.path.isfile(file_path):
        if raise_if_not_found:
            raise FileNotFoundError(f"The file '{file_path}' does not exist.")
        
        return False
    
    return True

def main():    
    try:
        args = parse_arguments()
        
        file_name = args.FileName
        history = int(args.History)
        delimeter = args.Delimeter
        closed_date_column = args.ClosedDateColumn
        date_format = args.DateFormat    
        
        remaining_items = int(args.RemainingItems)
        
        target_date = (datetime.now() + timedelta(days=14)).date()
        
        if args.TargetDate:        
            target_date = datetime.strptime(args.TargetDate, args.TargetDateFormat).date()
    
        csv_service = CsvService()
        monte_carlo_service = MonteCarloService(history, args.SaveCharts)

        print_logo()
        
        print("================================================================")
        print("Starting montecarlocsv with following Parameters")
        print("================================================================")  
        print("FileName: {0}".format(file_name))
        print("Delimeter: {0}".format(delimeter))
        print("ClosedDateColumn: {0}".format(closed_date_column))
        print("DateFormat: {0}".format(date_format))
        print("TargetDate: {0}".format(target_date))
        print("History: {0}".format(history))        
        print("Save Charts: {0}".format(args.SaveCharts))
        print("----------------------------------------------------------------")
                
        if file_name == '':
            print("No csv file specified - generating example file with random values")
            file_name = os.path.join(os.getcwd(), "ExampleFile.csv")
            if not check_if_file_exists(file_name):
                csv_service.write_example_file(file_name, delimeter, closed_date_column, history, date_format)
            

        closed_items_history = get_closed_items_history(csv_service, monte_carlo_service, file_name, delimeter, closed_date_column, date_format)        
        if len(closed_items_history) < 1:
            print("No closed items - skipping prediction")
            exit()

        ## Run How Many Predictions via Monte Carlo Simulation for our specified target date
        predictions_howmany_50 = predictions_howmany_70 = predictions_howmany_85 = predictions_howmany_95 = 0
        if target_date:
            (predictions_howmany_50, predictions_howmany_70, predictions_howmany_85, predictions_howmany_95) = monte_carlo_service.how_many(target_date, closed_items_history)       


        ## Run When Predictions via Monte Carlo Simulation - only possible if we have specified how many items are remaining
        predictions_when_50 = predictions_when_70 = predictions_when_85 = predictions_when_95 = datetime.today()
        predictions_targetdate_likelyhood = None

        if remaining_items > 0:
            (predictions_when_50, predictions_when_70, predictions_when_85, predictions_when_95, predictions_targetdate_likelyhood) = monte_carlo_service.when(remaining_items, closed_items_history, target_date)

            
        print("================================================================")
        print("Summary")
        print("================================================================")

        print("How many items will be done by {0}:".format(target_date))
        print("50%: {0}".format(predictions_howmany_50))
        print("70%: {0}".format(predictions_howmany_70))
        print("85%: {0}".format(predictions_howmany_85))
        print("95%: {0}".format(predictions_howmany_95))
        print("----------------------------------------")

        if remaining_items != 0:
            print("When will {0} items be done:".format(remaining_items))
            print("50%: {0}".format(predictions_when_50))
            print("70%: {0}".format(predictions_when_70))
            print("85%: {0}".format(predictions_when_85))
            print("95%: {0}".format(predictions_when_95))
            print("----------------------------------------")
            print("Chance of finishing the {0} remaining items till {1}: {2}%".format(remaining_items, target_date, predictions_targetdate_likelyhood))
            
            
        print()
        print()
        print()
        print("🛈 Want to learn more about how all of this works? Check out out website! 🛈")
        print("🔗 https://letpeople.work 🔗")
        
    except Exception as exception:
        print("Error while executing montecarloscsv:")
        print(exception)
        
        print("🪲 If the problem cannot be solved, consider opening an issue on GitHub: https://github.com/LetPeopleWork/MonteCarloCSV/issues 🪲")

if __name__ == "__main__":    
    main()


from .WorkItem import WorkItem
from datetime import datetime, timedelta
import random
import csv

class CsvService:    
       
    def get_closed_items(self, file_path, delimeter, column_name, date_format):
        print("Loading Items from CSV File: '{0}'. Column Name '{1}' and Date Format '{2}'".format(file_path, column_name, date_format))
        work_items = []
        
        with open(file_path, 'r', encoding='utf-8-sig') as file:
            csv_reader = csv.DictReader(file, delimiter=delimeter)
            
            for row in csv_reader:
                closed_date_raw = row[column_name]

                if closed_date_raw:
                    closed_date = datetime.strptime(closed_date_raw, date_format)           
                    work_items.append(WorkItem(closed_date))
        
        print("Found {0} Items in the CSV".format(len(work_items)))

        return work_items

    def write_example_file(self, file_path, delimeter, column_name, history, date_format):
        print("Writing Example File with random values to {0}".format(file_path))
        
        with open(file_path, 'w', newline='') as file:
            writer = csv.writer(file, delimiter=delimeter)
            field = [column_name]
            
            # Write Header
            writer.writerow(field)
            
            # Generate random entries
            for _ in range(30):
                random_days_ago = random.randint(0, history)
                random_date = datetime.now() - timedelta(days=random_days_ago)
                formatted_date = random_date.strftime(date_format)
                writer.writerow([formatted_date])


class WorkItem:
    def __init__(self, closed_date):
        self.closed_date = closed_date
            
    def to_dict(self):
            return {
                'closed_date': self.closed_date.date(),
            }


from setuptools import setup, find_packages

setup(
    name='montecarlocsv',
    version='0.9.5',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'MonteCarloCSV': ['logo.png'],
    },
    install_requires=[
        "argparse", "pandas", "numpy", "matplotlib"
    ],
    entry_points={
        'console_scripts': [
            'montecarlocsv=MonteCarloCSV.main:main',
        ],
    },
    author='Benjamin Huser-Berta',
    author_email='benj.huser@gmail.com',
    description='A package to generate Monte Carlo Simulations from CSV files.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://letpeople.work',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
)



import random
from datetime import date, datetime, timedelta

import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from matplotlib.offsetbox import OffsetImage, AnnotationBbox

import pandas as pd

import os

class MonteCarloService:
    
    def __init__(self, history_in_days, save_charts = False, trials=10000):
        self.trials = trials        
        self.history_in_days = history_in_days
        
        self.percentile_50 = 0.5
        self.percentile_70 = 0.7
        self.percentile_85 = 0.85
        self.percentile_95 = 0.95

        # Save the plot as an image in the "Charts" folder next to the script
        self.charts_folder = os.path.join(os.getcwd(), 'Charts')
        self.save_charts = save_charts

        self.current_date = datetime.now().strftime('%d.%m.%Y')
        
        script_dir = os.path.dirname(os.path.abspath(__file__))
        logo_path = os.path.join(script_dir, "logo.png")
        self.logo = mpimg.imread(logo_path)
        
        if save_charts and not os.path.exists(self.charts_folder):
            os.makedirs(self.charts_folder)
        
    def create_closed_items_history(self, items):
        print("Getting items that were done in the last {0} days...".format(self.history_in_days))       
        time_delta = date.today() - timedelta(self.history_in_days)
        df = pd.DataFrame.from_records([item.to_dict() for item in items])        
        
        closed_items = pd.DataFrame()
        
        closed_items = pd.concat([closed_items, df[df.closed_date >= time_delta]])
        
        closed_items_hist = closed_items["closed_date"].value_counts().to_dict()

        if self.save_charts:
            tp_run_chart_path = os.path.join(self.charts_folder, 'Throughput_Run_Chart.png')
            plt.figure(figsize=(15, 9))
            
            print("Storing Chart at {0}".format(tp_run_chart_path))
            plt.bar(list(closed_items_hist.keys()), closed_items_hist.values(), color='b')
            
            self.add_timestamp(plt)
            self.add_logo(plt)
            
            plt.savefig(tp_run_chart_path)
        
        print("Found {0} items that were closed in the last {1} days".format(len(closed_items), self.history_in_days))
                    
        return closed_items_hist
    
    def how_many(self, target_date, closed_items_history):
        print("--------------------------------")
        print("Running Monte Carlo Simulation - How Many items will be done till {0}".format(target_date))
        print("--------------------------------")
        monte_carlo_simulation_results = self.__run_monte_carlo_how_many(target_date, closed_items_history)
        
        return self.__get_predictions_howmany(monte_carlo_simulation_results)
        
    def when(self, remaining_items, closed_items_history, target_date = None):
        print("--------------------------------")
        print("Running Monte Carlo Simulation - When will {0} items be done".format(remaining_items))
        print("--------------------------------")
        monte_carlo_simulation_results = self.__run_monte_carlo_when(remaining_items, closed_items_history)
        
        days_to_target_date = None
        if target_date:
            days_to_target_date = (target_date - date.today()).days
            print("{0} days to target date".format(days_to_target_date))
        
        (predicted_date_50, predicted_70, predicted_date_85, predicted_date_95, target_date_likelyhood) = self.__get_predictions_when(monte_carlo_simulation_results, days_to_target_date)
        return (predicted_date_50, predicted_70, predicted_date_85, predicted_date_95, target_date_likelyhood)

    def __run_monte_carlo_when(self, remaining_items, closed_items_history):        
        time_delta = date.today() - timedelta(self.history_in_days)
        monte_carlo_data = self.__prepare_monte_carlo_dataset(time_delta, closed_items_history)            
                
        mc_results = {}
                
        for i in range(self.trials):
            day_count = 0
            finished_item_count = 0
                    
            while finished_item_count < remaining_items:
                day_count += 1
                rand = random.randint(0, self.history_in_days - 1)
                finished_item_count += monte_carlo_data[rand]
                        
            if day_count in mc_results:
                mc_results[day_count] += 1
            else:
                mc_results[day_count] = 1
                
        return mc_results

    def __get_predictions_when(self, mc_results, days_to_target_date = None):        
        sorted_dict = {k:v for k,v in sorted(mc_results.items())}
                
        percentile_50_target = self.trials * 0.5
        percentile_70_target = self.trials * 0.7
        percentile_85_target = self.trials * 0.85
        percentile_95_target = self.trials * 0.95


        percentile_50 = 0
        percentile_70 = 0
        percentile_85 = 0
        percentile_95 = 0
        
        # Initialize trials in time with the number of trials - if we have 100% chance of hitting the target date, we don't need to calculate it
        trials_in_time = self.trials if days_to_target_date else -1
        
        count = 0
                    
        for key in sorted_dict:
            value = sorted_dict[key]

            count += value
                    
            if (percentile_50 == 0 and count >= percentile_50_target):
                percentile_50 = key
            elif (percentile_70 == 0 and count >= percentile_70_target):
                percentile_70 = key
            elif (percentile_85 == 0 and count >= percentile_85_target):
                percentile_85 = key
            elif (percentile_95 == 0 and count >= percentile_95_target):
                percentile_95 = key
                
            if trials_in_time == self.trials and key >= days_to_target_date:
                trials_in_time = count
                
        
        predicted_date_50 = date.today() + timedelta(percentile_50)
        predicted_date_70 = date.today() + timedelta(percentile_70)
        predicted_date_85 = date.today() + timedelta(percentile_85)
        predicted_date_95 = date.today() + timedelta(percentile_95)

        print("50 Percentile: {0} days - Predicted Date: {1}".format(percentile_50, predicted_date_50))
        print("70 Percentile: {0} days - Predicted Date: {1}".format(percentile_70, predicted_date_70))
        print("85 Percentile: {0} days - Predicted Date: {1}".format(percentile_85, predicted_date_85))
        print("95 Percentile: {0} days - Predicted Date: {1}".format(percentile_95, predicted_date_95))
        
        prediction_targetdate = 0
        if days_to_target_date:
            prediction_targetdate = (100 / self.trials) * trials_in_time
            print("Chance of hitting target date: {0}".format(prediction_targetdate))
        
        if self.save_charts:
            vertical_lines_data = [(percentile_50, '50th Percentile', "red"), (percentile_70, '70th Percentile', "orange"), (percentile_85, '85th Percentile', "lightgreen"), (percentile_95, '95th Percentile', "darkgreen")] 
            plt.figure(figsize=(15, 9))
            when_chart_path = os.path.join(self.charts_folder, 'MC_When.png')
            print("Storing Chart at {0}".format(when_chart_path))
            plt.bar(list(sorted_dict.keys()), sorted_dict.values(), color='g')

            for position, line_name, color in vertical_lines_data:
                plt.axvline(x=position, color=color, linestyle='--', label="{0} ({1})".format(line_name, position))

            plt.legend()
            
            self.add_timestamp(plt)
            self.add_logo(plt)

            plt.savefig(when_chart_path)

        return (predicted_date_50, predicted_date_70, predicted_date_85, predicted_date_95, prediction_targetdate)

    def __run_monte_carlo_how_many(self, prediction_date, closed_items_hist):
        time_delta = date.today() - timedelta(self.history_in_days)
        monte_carlo_data = self.__prepare_monte_carlo_dataset(time_delta, closed_items_hist)            
                
        mc_results = {}
        amount_of_days = (prediction_date - date.today()).days
                
        for i in range(self.trials):
            day_count = 0
            finished_item_count = 0
                    
            while day_count < amount_of_days:
                day_count += 1
                rand = random.randint(0, self.history_in_days - 1)
                finished_item_count += monte_carlo_data[rand]
                        
            if finished_item_count in mc_results:
                mc_results[finished_item_count] += 1
            else:
                mc_results[finished_item_count] = 1
                
        return mc_results

    def __get_predictions_howmany(self, mc_results):        
        sorted_dict = {k:v for k,v in sorted(mc_results.items(), reverse=True)}
                
        percentile_50_target = self.trials * self.percentile_50
        percentile_70_target = self.trials * self.percentile_70
        percentile_85_target = self.trials * self.percentile_85
        percentile_95_target = self.trials * self.percentile_95
        
        percentile_50 = 0
        percentile_70 = 0
        percentile_85 = 0
        percentile_95 = 0
        
        count = 0
                
        for key in sorted_dict:
            value = sorted_dict[key]

            count += value
                    
            if (percentile_50 == 0 and count >= percentile_50_target):
                percentile_50 = key
            elif (percentile_70 == 0 and count >= percentile_70_target):
                percentile_70 = key
            elif (percentile_85 == 0 and count >= percentile_85_target):
                percentile_85 = key
            elif (percentile_95 == 0 and count >= percentile_95_target):
                percentile_95 = key
                
                
        print("50 Percentile: {0} Items".format(percentile_50))
        print("70 Percentile: {0} Items".format(percentile_70))
        print("85 Percentile: {0} Items".format(percentile_85))
        print("95 Percentile: {0} Items".format(percentile_95))

        if self.save_charts:
            vertical_lines_data = [(percentile_50, '50th Percentile', "red"), (percentile_70, '70th Percentile', "orange"), (percentile_85, '85th Percentile', "lightgreen"), (percentile_95, '95th Percentile', "darkgreen")] 

            plt.figure(figsize=(15, 9))
            how_many_chart_path = os.path.join(self.charts_folder, 'MC_HowMany.png')
            print("Storing Chart at {0}".format(how_many_chart_path))
            plt.bar(list(sorted_dict.keys()), sorted_dict.values(), color='g')
            
            for position, line_name, color in vertical_lines_data:
                plt.axvline(x=position, color=color, linestyle='--', label="{0} ({1})".format(line_name, position))

            plt.legend()
            
            self.add_timestamp(plt)
            self.add_logo(plt)

            plt.savefig(how_many_chart_path)
        
        return (percentile_50, percentile_70, percentile_85, percentile_95)

    def __prepare_monte_carlo_dataset(self, time_delta, closed_items_hist):
        monte_carlo_data = {}
        
        for i in range((date.today() - time_delta).days):
            day = date.today() - timedelta(days=i)
            monte_carlo_data[i] = 0

            if day in closed_items_hist:
                monte_carlo_data[i] = closed_items_hist[day]
                
        
        return monte_carlo_data       
    
    def add_timestamp(self, plt):
        plt.text(1, 1.02, f"Generated on {self.current_date}", transform=plt.gca().transAxes, fontsize=10, ha='right', va='top')
    
    def add_logo(self, plt):
        # Add logo
        imagebox = OffsetImage(self.logo, zoom=0.48)
        ab = AnnotationBbox(imagebox, (0.065, 1.08), xycoords='axes fraction', frameon=False, box_alignment=(1, 1))
        plt.gca().add_artist(ab)